@@ -5,7 +5,7 @@
 **
 ** Comments: 
 **
-** $Id: gre.c,v 1.5 2001/01/20 00:30:31 mborella Exp $
+** $Id: gre.c,v 1.8 2001/04/02 17:41:01 mborella Exp $
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
@@ -42,10 +42,16 @@ extern strmap_t ethertype_map[];
 
 void dump_gre(packet_t *pkt)
 {
-  gre_v0_t gre0;
+  gre_v0_rfc2784_t gre0;
   gre_v1_t gre1;
   ethertype_func_t et;
 
+  /*
+   * Stats accounting
+   */
+
+  stats_update(STATS_GRE);
+
   /*
    * Look at the first 2 bytes of the GRE header so that we can check the 
    * version
@@ -63,71 +69,217 @@ void dump_gre(packet_t *pkt)
     case 0:
       {
 	u_int16_t reserved;
-	u_int8_t  C;
+	u_int16_t protocol;
 	ethertype_func_t et;
 
-	/*
-	 * Get the header 
-	 */
-	
-	if (get_packet_bytes((u_int8_t *) &gre0, pkt, sizeof(gre_v0_t)) == 0)
-	  return;
-	
-	/*
-	 * Conversions
+	/* 
+	 * Please note: If the reserved field is 0, we assume that 
+	 * the GRE packet follows RFC2784 and act accordingly.  Otherwise
+	 * we assume the RFC1701 interpretation.  If key and sequence
+	 * number fields are used, ala RFC2890, then we parse according
+	 * RFC1701.  Got all that?
 	 */
 
 	reserved = (gre0.reserved_high << 5) || gre0.reserved_low;
-	gre0.protocol = ntohs(gre0.protocol);
-	C = gre0.c;
+	protocol = gre0.protocol;
+	if (reserved == 0)
+	  { 
+	    gre_v0_rfc2784_t gre;
+	    u_int8_t C;
 
-	/* Dump the header */
-	if (!my_args->n)
-	  {
-	    if (my_args->m)
+	    /*
+	     * Get the header 
+	     */
+	
+	    if (get_packet_bytes((u_int8_t *) &gre, pkt, 
+				 sizeof(gre_v0_rfc2784_t)) == 0)
+	      return;
+	    
+	    /*
+	     * Conversions
+	     */
+	    
+	    gre.protocol = ntohs(gre.protocol);
+	    C = gre.c;
+	    
+	    /* Dump the header */
+	    if (!my_args->n)
 	      {
-		display_minimal_string("| GRE v0 ");
-		display_minimal((u_int8_t *) &gre0.protocol, 2, DISP_HEX);
+		if (my_args->m)
+		  {
+		    display_minimal_string("| GRE v0 ");
+		    display_minimal((u_int8_t *) &gre.protocol, 2, DISP_HEX);
+		  }
+		else
+		  {
+		    display_header_banner("GREv0 Header");
+		    display("C", &C, 1, DISP_DEC);
+		    display("Reserved", (u_int8_t *) &reserved, 2, DISP_DEC);
+		    display_strmap_hex("Protocol", gre.protocol, 
+				       ethertype_map);
+		  }
 	      }
-	    else
+	    
+	    /*
+	     * Get the checksum and reserved fields, if they exist
+	     */
+	    
+	    if (C)
 	      {
-		display("C", &C, 1, DISP_DEC);
-		display("Reserved", (u_int8_t *) &reserved, 2, DISP_DEC);
-		display("Protocol", (u_int8_t *) &gre0.protocol, 2, DISP_HEX);
+		u_int16_t checksum;
+		u_int16_t reserved;
+		
+		/* Get the fields */
+		if (get_packet_bytes((u_int8_t *) &checksum, pkt, 2) == 0)
+		  return;
+		if (get_packet_bytes((u_int8_t *) &reserved, pkt, 2) == 0)
+		  return;
+		
+		/* conversions */
+		checksum = ntohs(checksum);
+		reserved = ntohs(reserved);
+		
+		if (!my_args->m)
+		  {
+		    display("Checksum", (u_int8_t *) &checksum, 2, DISP_DEC);
+		    display("Reserved", (u_int8_t *) &reserved, 2, DISP_DEC);
+		  }
 	      }
 	  }
-
-	/*
-	 * Get the checksum and reserved fields, if they exist
-	 */
-
-	if (C)
+	else
 	  {
-	    u_int16_t checksum;
-	    u_int16_t reserved;
+	    gre_v0_rfc1701_t gre;
+	    u_int8_t C, R, K, S, s, recur, flags, version;
 
-	    /* Get the fields */
-	    if (get_packet_bytes((u_int8_t *) &checksum, pkt, 2) == 0)
+	    /*
+	     * Get the header 
+	     */
+	
+	    if (get_packet_bytes((u_int8_t *) &gre, pkt, 
+				 sizeof(gre_v0_rfc1701_t)) == 0)
 	      return;
-	    if (get_packet_bytes((u_int8_t *) &reserved, pkt, 2) == 0)
-	      return;
-
-	    /* conversions */
-	    checksum = ntohs(checksum);
-	    reserved = ntohs(reserved);
-
-	    if (!my_args->m)
+	    
+	    /*
+	     * Conversions
+	     */
+	    
+	    gre.protocol = ntohs(gre.protocol);
+	    C = gre.c;
+	    R = gre.r;
+	    K = gre.k;
+	    S = gre.S;
+	    s = gre.s;
+	    recur = gre.recur;
+	    flags = gre.flags;
+	    version = gre.version;
+	    
+	    /* Dump the header */
+	    if (!my_args->n)
 	      {
-		display("Checksum", (u_int8_t *) &checksum, 2, DISP_DEC);
-		display("Reserved", (u_int8_t *) &reserved, 2, DISP_DEC);
+		if (my_args->m)
+		  {
+		    display_minimal_string("| GRE v0 ");
+		    display_minimal((u_int8_t *) &gre.protocol, 2, DISP_HEX);
+		  }
+		else
+		  {
+		    display_header_banner("GREv0 Header");
+		    display("C", &C, 1, DISP_DEC);
+		    display("R", &R, 1, DISP_DEC);
+		    display("K", &K, 1, DISP_DEC);
+		    display("S", &S, 1, DISP_DEC);
+		    display("s", &s, 1, DISP_DEC);
+		    display("Recur", &recur, 1, DISP_DEC);
+		    display("Flags", &flags, 1, DISP_DEC);
+		    display("Version", &version, 1, DISP_DEC);
+		    display_strmap_hex("Protocol", gre.protocol, 
+				       ethertype_map);
+		  }
+	      }
+
+	    /* 
+	     * Parse options fields. Note that if the C or the R bits are
+	     * present then both the checksum and offset fields are present
+	     */
+
+	    if (C || R)
+	      {
+		if (!my_args->n)
+		  {
+		    u_int16_t checksum;
+		    u_int16_t offset;
+
+		    /* 
+		     * Nothing to report for minimal mode here, so just get
+		     * the checksum and offset and display them.
+		     */
+
+		    if (get_packet_bytes((u_int8_t *) &checksum, pkt, 2)
+			== 0)
+		      return;
+		    checksum = ntohs(checksum);
+		    display("Checksum", (u_int8_t *) &checksum, 2, DISP_DEC);
+
+		    if (get_packet_bytes((u_int8_t *) &offset, pkt, 2)
+			== 0)
+		      return;
+		    offset = ntohs(offset);
+		    display("Offset", (u_int8_t *) &offset, 2, DISP_DEC);
+		  }
+	      }
+
+	    /* Check for the K bit (key field) */
+	    if (K)
+	      {
+		if (!my_args->n)
+		  {
+		    u_int32_t key;
+
+		    if (get_packet_bytes((u_int8_t *) &key, pkt, 4) == 0)
+		      return;
+		    key = ntohl(key);
+		    display("Key", (u_int8_t *) &key, 4, DISP_DEC);
+		  }
+	      }
+
+	    /* Check for the sequence number field */
+	    if (S)
+	      {
+		if (!my_args->n)
+		  {
+		    u_int32_t seqno;
+
+		    if (get_packet_bytes((u_int8_t *) &seqno, pkt, 4) == 0)
+		      return;
+		    seqno = ntohl(seqno);
+		    display("Sequence number", (u_int8_t *) &seqno, 4, 
+			    DISP_DEC);
+		  }
+	      }
+	    
+	    /* Check for routing field */
+	    if (R)
+	      {
+		if (!my_args->n)
+		  {
+		    u_int32_t routing;
+
+		    if (get_packet_bytes((u_int8_t *) &routing, pkt, 4) == 0)
+		      return;
+		    routing = ntohs(routing);
+		    display("Routing", (u_int8_t *) &routing, 4, DISP_DEC);
+		  }
 	      }
 	  }
+	
+	/* flush the hexbuffer */
+	hexbuffer_flush();
 
 	/*
 	 * Display the rest of the packet
 	 */
 	
-	et = ethertype2func(gre0.protocol);
+	et = ethertype2func(protocol);
 	if (et)
 	  {
 	    et(pkt);
@@ -172,12 +324,12 @@ void dump_gre(packet_t *pkt)
 	  {
 	    if (my_args->m)
 	      {
-		display_minimal_string("GREv1 ");
+		display_minimal_string("| GREv1 ");
 	      }
 	    else
 	      {
 		/* announcement */
-		display_header_banner("GRE Header");
+		display_header_banner("GREv1 Header");
 		
 		/* fields */
 		display("Checksum (C) bit", &C, 1, DISP_BINNLZ);
